/*
   Copyright (C) 2002   Alexandre Courbot
   Copyright (C) 2002   Dmitri Katchalov

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.

   See the COPYING file for more details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/io.h> 
#include <sys/time.h> 
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>
#include <linux/parport.h>
#include "rome.h"

/* Some convenience / portability macros */
#define INB(P, B)   B = inb (P)
#define OUTB(P, B)    outb (B, P)

/* Number of nanoseconds to wait for the device to answer before
   considering a timeout */ 
#define DATA_TIMEOUT 1000000


/* Hopefully, the base adress of LPT1 */ 
#define PORT_BASE 0x378

/* Offsets from the base port adress to port registers */ 
typedef enum
{ DATA_REG = 0, STATUS_REG = 1, CONTROL_REG = 2 } PARPORT_REGS; 


static rome_error re;

#define STOP_ON_FAILURE(X) if ((re = X) != success) return re


rome_error rome_filelist_init (rome_filelist * flist) 
{
    flist->nbr_files = 0;

    return success; 
}

rome_error rome_filelist_cleanup (rome_filelist * flist) 
{
    int i;

    for (i = 0; i < flist->nbr_files; i++) 
    {
        free (flist->file[i].filename); 
    }
    
    rome_filelist_init (flist); 

    return success; 
}

rome_error rome_filelist_add (rome_filelist * flist, char * filename, int size,
                              char * date, char * rpacket) 
{
    flist->file[flist->nbr_files].filename = strdup (filename); 
    flist->file[flist->nbr_files].size = size; 
    strcpy (flist->file[flist->nbr_files].date, date);
    memcpy (flist->file[flist->nbr_files].rawpacket, rpacket, 152); 

    flist->nbr_files++; 
    
    return success; 
}
     
/* Global variables, used to keep the port state before initialisation
   so it can be restored at exit */ 
static char orig_status, orig_control; 


static unsigned short int data_reg; 
static unsigned short int status_reg; 
static unsigned short int control_reg; 

static unsigned char prev_clk;
static unsigned char prev_ack; 

static int read_buf_size; 
static unsigned char read_buf[70000]; 

/* Prints a byte in binary (dirty) */
void print_byte (char b) 
{
    printf ("%d%d%d%d%d%d%d%d", (b & 128) != 0, (b & 64) != 0,
            (b & 32) != 0, (b & 16) != 0, (b & 8) != 0, (b & 4) != 0,
            (b & 2) != 0, (b & 1) != 0); 
}


void timeval_add_usec (struct timeval * tv, int val) 
{
    tv->tv_usec += val;
    if (tv->tv_usec >= 1000000) 
    {
        tv->tv_sec++;
        tv->tv_usec -= 1000000; 
    }     
}

int timeval_diff (struct timeval * tv1, struct timeval * tv2) 
{
    return (tv1->tv_sec - tv2->tv_sec) * 1000000 +
            tv1->tv_usec - tv2->tv_usec;
}
 
rome_error rome_init () 
{
    unsigned char c;
    
    data_reg = PORT_BASE + DATA_REG; 
    status_reg = PORT_BASE + STATUS_REG; 
    control_reg = PORT_BASE + CONTROL_REG;
    
    printf ("Getting extra permissions for ports access...   ");  
    if (iopl (3)) 
    {
        printf ("failed!\n"); 
        perror ("iopl");
        printf ("Are you running the program as root?\n"); 
        return failure; 
    }     
    printf ("Ok!\n");

    /* Saves the ports status into these variables */ 
    INB (status_reg, orig_status);
    INB (control_reg, orig_control);

    INB (control_reg, c);
    c &= (~reverse);
    c &= (~nreverserequest); 
    c &= (~irqen);
    c |= (active1284);
    
    OUTB (control_reg, c);
    
    INB (status_reg, c); 
    c &= periphclk; 
    prev_clk = c;
    
    INB (status_reg, c);
    c &= periphack;
    prev_ack = c;

    INB (status_reg, c);
    INB (control_reg, c);
    
    return success; 
}

rome_error rome_cleanup () 
{
    OUTB (control_reg, orig_control);
    OUTB (status_reg, orig_status);

    return success; 
}

rome_error rome_set_forward () 
{
    unsigned char c;

    /* Set Reverse to LOW */ 
    INB (control_reg, c);
    OUTB (control_reg, c & ~reverse); 

    /* Set nReverseRequest to LOW */ 
    INB (control_reg, c);
    OUTB (control_reg, c & ~nreverserequest); 

    return success; 
}

rome_error rome_set_reverse () 
{
    unsigned char c;
    
    /* Set Reverse to HIGH */ 
    INB (control_reg, c);
    OUTB (control_reg, c | reverse); 

    /* Set nReverseRequest to HIGH */ 
    INB (control_reg, c);
    OUTB (control_reg, c | nreverserequest); 

    return success; 
}
 
rome_error rome_wait_for (unsigned char bit, unsigned char value) 
{
    struct timeval ack_timeout, t;
    unsigned char x = value ? bit : 0;
    int diff; 
    unsigned char c; 
    
    /* Get the current time and add the timeout to it */ 
    gettimeofday (&ack_timeout, NULL);
    timeval_add_usec (&ack_timeout, DATA_TIMEOUT); 

    while (1) 
    {
        /* Check whether the bit has the expected value */ 
        INB (status_reg, c);
        if ((c & bit) == x) break;

        /* Else check whether we reached the timeout -> failure. */
        gettimeofday (&t, NULL);
        diff = timeval_diff (&ack_timeout, &t);
        if (diff <= 0) return timeout;
    }

    return success; 
}
 
rome_error rome_write_buf (char * buf, int size) 
{     
    int currentbyte;
    unsigned char c;

    /* Set forward mode */ 
    STOP_ON_FAILURE (rome_set_forward ());

    /* Wait for nackreverse to go high (the device acknowledges the forward
       transfert request). */
    STOP_ON_FAILURE (rome_wait_for (nackreverse, 1)); 
    
    /* Get the current state of periphack, for later */ 
    INB (status_reg, c); 
    prev_ack = c & periphack;

    for (currentbyte = 0; currentbyte < size; currentbyte++) 
    { 
        /* Write the data byte on the data register */ 
        OUTB (data_reg, buf[currentbyte]);
         
        /* Toggle HostClk to warn the device there is data
           available */ 
        INB (control_reg, c); 
        OUTB (control_reg, c ^ hostclk);
        
        /* Now wait the device to acknowledge the data
           (periphack should toggle) */
        STOP_ON_FAILURE (rome_wait_for (periphack, !prev_ack)); 
        
        /* The data has been received, good. Now update prev_ack
           and increase the counter */
        prev_ack = !prev_ack;
    }
    
    return success; 
}

int rome_read_buf (int count) 
{
    int currentbyte;
    unsigned char c;
    
    for (currentbyte = 0; currentbyte < count; currentbyte++)
    { 
        /* Wait for periphclk to toggle to indicate a new byte
           is available */
        if ((re = rome_wait_for (periphclk, !prev_clk)) != success) return 0; 
        
        prev_clk = !prev_clk;
        
        /* Read the byte */ 
        INB (data_reg, read_buf[read_buf_size]);

        read_buf_size++; 
        
        INB (control_reg, c);
        OUTB (control_reg, c ^ hostack); 
    } 
    
    return currentbyte; 
}

rome_error rome_read_packet () 
{
    unsigned short int size; 
    unsigned char checksum; 
    
    /* Set reverse mode */ 
    STOP_ON_FAILURE (rome_set_reverse ()); 
    
    /* Wait for nAckReverse to go low (the device acknowledges the reverse
       transfert request). */
    STOP_ON_FAILURE (rome_wait_for (nackreverse, 0)); 

    /* Start reading at the beginning of the buffer */ 
    read_buf_size = 0;

    /* Attempt to read the first byte (version number) */
    if (rome_read_buf (1) != 1) return failure; 

    /* Version number should always be 2 */
    if (read_buf[0] != 2) return bad_version; 

    /* Reset the buffer, as we don't care to have the version
       number into it */
    read_buf_size = 0; 

    /* Attempt to read the size of the data following */
    if (rome_read_buf (2) != 2) return failure; 
 
    /* Convert the size to a short int */ 
    size = (unsigned short int) read_buf[1] << 8;
    size += read_buf[0];

    /* Read the actual data + the 2 bytes checksum */ 
    if (rome_read_buf (size + 2) != size + 2) return failure;

    /* Now calculate the checksum */ 
    /* Perform a XOR on the two size bytes */
    checksum = read_buf[0] ^ read_buf[1];

    /* If we read more than the checksum */
    if (size + 2 >= 2) 
    {
        int cpt = 2;
        do checksum ^= read_buf[cpt++]; while (cpt <= size + 2); 
    }

    if (checksum != read_buf[size + 2 + 1]) return checksum_error;
     
    return success; 
}

rome_error rome_get_info (rome_info * ifo) 
{ 
    /* Get player info command */ 
    unsigned char command[9] = { 0x02, 0x04, 0x00, 0x10, 0x01, 0x00, 0x00, 0x03, 0x16 }; 
    
    STOP_ON_FAILURE (rome_write_buf (command, sizeof (command)));  
    
    STOP_ON_FAILURE (rome_read_packet ()); 
    
    if (read_buf[2] != 0x10) return bad_version;
    if (read_buf[3] != 1) return bad_version; 

    ifo->mem_tot = read_buf[10] + (read_buf[11] << 8) +
        (read_buf[12] << 16) + (read_buf[13] << 24); 

    ifo->mem_avail = read_buf[14] + (read_buf[15] << 8) +
        (read_buf[16] << 16) + (read_buf[17] << 24); 

    ifo->nbr_files = read_buf[18] + (read_buf[19] << 8) +
        (read_buf[20] << 16) + (read_buf[21] << 24); 
    
    return success; 
}
 
rome_error rome_get_serial_number () 
{
    /* Get serial number request */
    unsigned char command[9] = { 0x02, 0x04, 0x00, 0x10, 0x03, 0x00, 0x00, 0x03, 0x14 };

    STOP_ON_FAILURE (rome_write_buf (command, sizeof (command))); 

    STOP_ON_FAILURE (rome_read_packet ()); 
         
    return success; 
}

rome_error rome_list_next_file (char * last) 
{
    /* Get file list command */ 
    unsigned char command[9] = { 0x02, 0x04, 0x00, 0x20, 0x04, 0x00, 0x00, 0x03, 0x23 }; 

    STOP_ON_FAILURE (rome_write_buf (command, sizeof (command))); 

    STOP_ON_FAILURE (rome_read_packet ()); 

    if (last) *last = read_buf[4]; 
    
    return success; 
}

rome_error rome_list_files (rome_filelist * flist) 
{
    char last;

    do
    {
        STOP_ON_FAILURE (rome_list_next_file (&last));
        rome_filelist_add (flist, read_buf + 10, (read_buf[9] << 24) +
                           (read_buf[8] << 16) + (read_buf[7] << 8) +
                           read_buf[6], read_buf + 138, read_buf + 6); 
    }while (last); 

    return success; 
}
 
rome_error rome_get_file (FILE * file, rome_filelist * flist, unsigned int nbr) 
{
    int outputsize;
    int i;
    unsigned char c = 0;
    unsigned int packetsize; 
    unsigned int bytesread = 0; 
    
    /*  Get file command */ 
    unsigned char init_command[161] = { 0x02, 0x9C, 0x00, 0x20, 0x02, 0x00, 0x00 }; 

    /* Download file packet command */ 
    unsigned char recv_command[9] = { 0x02, 0x04, 0x00, 0x20, 0x02, 0x00, 0x00, 0x03, 0x25 }; 

    /* Request file */
    /* Copy file identifier into the packet */ 
    memcpy (init_command + 7, flist->file[nbr].rawpacket, 152); 
    init_command[159] = 0x3;
    
    /* Calculate checksum */ 
    for (i = 1; i < sizeof (init_command) - 1; i++) 
        c ^= init_command[i]; 

    init_command[160] = c; 

    /* Write request packet */ 
    STOP_ON_FAILURE (rome_write_buf (init_command, sizeof (init_command)));
 
    /* Receive answer packet */ 
    STOP_ON_FAILURE (rome_read_packet ());

    /* Check whether our demand is validated */
    if (read_buf[2] != 0x20 || read_buf[3] != 0x02) return failure; 
    
    
    /* Ok, now get the file data itself */
    outputsize = 0; 
    while (1) 
    {
        unsigned char notlastpacket;
        int i; 
         
        /* Send request for next packet */
        STOP_ON_FAILURE (rome_write_buf (recv_command, sizeof (recv_command)));
        
        /* Get next packet */
        STOP_ON_FAILURE (rome_read_packet ()); 
        
        packetsize = (read_buf[1] << 8) + read_buf[0];
        packetsize -= 4;

        notlastpacket = read_buf[4]; 
        
        bytesread += packetsize;
        
        fwrite (read_buf + 6, 1, packetsize, file); 
        
        for (i = 0; i < outputsize; i++) printf ("\b");  
        outputsize = printf ("%d bytes read", bytesread);
        fflush (stdout); 
        
        if (!notlastpacket) break; 
    } 

    printf ("\n"); 
    return success; 
}

rome_error rome_put_file (FILE * file, char * fname) 
{
    int outputsize;
    
    time_t currenttime;
    struct tm * timep; 
    int i;
    unsigned char c = 0;
    unsigned int packetsize; 
    unsigned int byteswritten = 0; 
    int filesize; 
    
    /*  Get file command */ 
    unsigned char init_command[161] = { 0x02, 0x9C, 0x00, 0x20, 0x01, 0x00, 0x00 }; 

    /* Upload file packet command */ 
    unsigned char send_command[65535] = { 0x02, 0x00, 0x00, 0x20, 0x01, 0x00, 0x00 }; 

    /* Request file */
    /* Create the file identifier */
    for (i = 7; i < 159; i++)
        init_command[i] = 0;

    /* Get the file size */
    fseek (file, 0, SEEK_END);
    filesize = ftell (file);
    fseek (file, 0, SEEK_SET); 

    init_command[7] = filesize & 0x000000FF; 
    init_command[8] = (filesize & 0x0000FF00) >> 8; 
    init_command[9] = (filesize & 0x00FF0000) >> 16; 
    init_command[10] = (filesize & 0xFF000000) >> 24; 
    
    /* Copy the filename */
    strcpy (init_command + 11, fname);

    /* Generate the date of creation */ 
    currenttime = time (NULL);
    timep = localtime (&currenttime);
    sprintf (init_command + 139, "%04d%02d%02d-%02d%02d%02d",
             1900 + timep->tm_year, timep->tm_mon + 1, timep->tm_mday,
             timep->tm_hour, timep->tm_min, timep->tm_sec);

    /* Calculate the checksum */
    init_command[159] = 0x03;
    
    for (i = 1; i < sizeof (init_command) - 1; i++) 
        c ^= init_command[i]; 

    init_command[160] = c; 

    STOP_ON_FAILURE (rome_write_buf (init_command, sizeof (init_command)));
    STOP_ON_FAILURE (rome_read_packet ()); 

    /* Check whether the received packet is correct */
    if (read_buf[2] != 0x20 || read_buf[3] != 0x01) return failure; 

    /* Ok, now we can actually send the file */
    outputsize = 0; 
    while (1) 
    {
        unsigned char notlastpacket; 
        int i; 

        /* Read next packet from file */
        packetsize = fread (send_command + 7, 1, 0xFE00, file);
        byteswritten += packetsize; 

        /* Write last packet info */ 
        if (byteswritten >= filesize) send_command[5] = 0;
        else send_command[5] = 1; 
        notlastpacket = send_command[5]; 
        
        /* Write the size of the packet into it */
        packetsize += 4; 
        send_command[1] = packetsize & 0xFF;
        send_command[2] = (packetsize & 0xFF00) >> 8; 
        
        /* Calculate the checksum */
        send_command[packetsize + 3] = 0x03;

        c = 0; 
        for (i = 1; i < packetsize + 4; i++) 
            c ^= send_command[i]; 
        
        send_command[packetsize + 4] = c;          

        /* Send the next packet */
        STOP_ON_FAILURE (rome_write_buf (send_command, packetsize + 5)); 
        
        /* Get ack packet */
        STOP_ON_FAILURE (rome_read_packet ()); 

        if (read_buf[2] != 0xFF || read_buf[3] != 0x01) return failure; 
        
        for (i = 0; i < outputsize; i++) printf ("\b");  
        outputsize = printf ("%d bytes written", byteswritten);
        fflush (stdout); 
        
        /* Last packet written? */ 
        if (!notlastpacket) break; 
    } 

    printf ("\n"); 
    
    
    return success; 
}

rome_error rome_remove_file (rome_filelist * flist, unsigned int nbr) 
{
    int i; 
    unsigned char c = 0; 
    /*  Remove file command */ 
    unsigned char command[161] = { 0x02, 0x9C, 0x00, 0x20, 0x03, 0x00, 0x00 }; 
    
    /* Request file */
    /* Copy file identifier into the packet */ 
    memcpy (command + 7, flist->file[nbr].rawpacket, 152); 
    command[159] = 0x3;
    
    /* Calculate checksum */ 
    for (i = 1; i < sizeof (command) - 1; i++) 
        c ^= command[i]; 

    command[160] = c; 

    /* Write request packet */ 
    STOP_ON_FAILURE (rome_write_buf (command, sizeof (command)));
 
    /* Receive answer packet */ 
    STOP_ON_FAILURE (rome_read_packet ());

    /* Check whether our demand is valid */
    if (read_buf[2] != 0xFF || read_buf[3] != 0x01) return failure; 
    
    return success; 
}
 
int main (int argc, char * argv[]) 
{     
    char letsexit = 0; 

    int i;
    unsigned short int size; 
    
    if (rome_init () != success)
    {
        printf ("Port initialisation failed!\n");
        return 1; 
    }
     
    printf ("Port initialisation successfull!\n"); 
     
    while (1) 
    {
        char c; 

        printf (" >> "); 

        c = getchar ();
        /* Get rid of the CR character */ 
        getchar (); 
        switch (c) 
        {
            case 'q':
                letsexit = 1;
                break; 

            case 's': 
                printf ("GetSerialNumber: %d\n", rome_get_serial_number ()); 
                size = (unsigned short int) read_buf[1] << 8;
                size += read_buf[0];
                for (i = 0; i < size + 2; i++) 
                    printf ("%u ", read_buf[i]); 
                printf ("\n"); 
                for (i = 0; i < size + 2; i++) 
                    printf ("%c ", read_buf[i]); 
                printf ("\n"); 
                break; 
                
            case 'i':
            {
                rome_info ifo;
                
                re = rome_get_info (&ifo);
                if (re != success) printf ("An error occured!\n");
                else printf ("Memory available/Total memory: %u/%u\n"
                             "Number of files on board: %d\n",
                             ifo.mem_avail, ifo.mem_tot, ifo.nbr_files); 
                break;
            }
            
            case 'l':
            {
                rome_filelist fl;
                int i; 

                rome_filelist_init (&fl);

                rome_list_files (&fl);
                
                for (i = 0; i < fl.nbr_files; i++)
                    printf ("%d  %d\t%s\t%s\n", i, fl.file[i].size, fl.file[i].filename,
                            fl.file[i].date); 

                rome_filelist_cleanup (&fl); 
                
                break;
            }

            case 'g':
            {
                unsigned int i; 
                rome_error c; 
                rome_filelist fl;
                FILE * file;

                rome_filelist_init (&fl); 
                rome_list_files (&fl);

                for (i = 0; i < fl.nbr_files; i++)
                    printf ("%d  %d\t%s\t%s\n", i, fl.file[i].size, fl.file[i].filename,
                            fl.file[i].date); 
                
                do
                {
                    printf ("File number to get: "); 
                    scanf ("%u", &i);
                }while (i > fl.nbr_files);
                
                file = fopen (fl.file[i].filename, "w");                 
                if (!file) 
                {
                    printf ("Cannot open %s!\n", fl.file[i].filename); 
                    rome_filelist_cleanup (&fl);
                    break;  
                }
                c = rome_get_file (file, &fl, i); 
                if (c != success) printf ("An error occured\n"); 
                else printf ("Successfull!\n"); 
                rome_filelist_cleanup (&fl);
                fclose (file); 
                break;
            }
            
            case 'p':
            {
                char fname[255]; 
                rome_error c;  
                FILE * file;
                int i, j; 
                
                printf ("File to send: ");
                fgets (fname, 255, stdin); 
                /* Remove the CR character */
                fname [strlen (fname) - 1] = 0; 
                
                file = fopen (fname, "r");
                if (!file) 
                {
                    printf ("Cannot open %s!\n", fname);
                    break;
                }

                /* Shorten the file name if necessary */
                i = strlen (fname);
                while (fname[i] != '/' && i) i--;
                i++; 
                j = 0; 
                while (fname[i]) fname[j++] = fname[i++];
                fname[j] = 0;
                printf ("%s\n", fname); 
                c = rome_put_file (file, fname);
                if (c != success) printf ("An error occured\n"); 
                else printf ("Successfull!\n"); 
                fclose (file); 
                break;
            }

            case 'r':
            {
                int i; 
                rome_error c; 
                rome_filelist fl;

                rome_filelist_init (&fl); 
                rome_list_files (&fl);
                
                for (i = 0; i < fl.nbr_files; i++)
                    printf ("%d  %d\t%s\t%s\n", i, fl.file[i].size, fl.file[i].filename,
                            fl.file[i].date); 
                
                do
                {
                    printf ("File number to remove (0-%d): ", fl.nbr_files - 1); 
                    scanf ("%u", &i);
                }while (i > fl.nbr_files);

                c = rome_remove_file (&fl, i);
                if (c != success) printf ("An error occured\n"); 
                else printf ("Successfull!\n"); 
                rome_filelist_cleanup (&fl);
                break; 
            }
            
            default:
                printf ("Unknown command\n");
                break; 
        }
                 
        if (letsexit) break;
        getchar ();  
    }
    
 
    rome_cleanup ();
    
    return EXIT_SUCCESS; 
}
